/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is the reusable ccl library
 * (http://www.kclee.com/clemens/java/ccl/).
 *
 * The Initial Developer of the Original Code is Chr. Clemens Lee.
 * Portions created by Chr. Clemens Lee are Copyright (C) 1999
 * Chr. Clemens Lee. All Rights Reserved.
 *
 * Contributor(s): Chr. Clemens Lee <clemens@kclee.com>
 */

package ccl.util;

import java.awt.Color;
import java.awt.SystemColor;
import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;

/**
 * Manages the ini file and configuration stuff for an
 * application.
 *
 * Right now localisation of ini file does not work with
 * src and classes directory separation.
 * When the ini file is inside an archive, user can't
 * change the ini file values.
 * In the future there should be one application ini file
 * that can be located like a resource and a user ini file
 * located in the user's home directory.
 *
 * @version $Id: Init.java,v 1.41 2000/05/14 14:30:35 clemens Exp clemens $
 * @author <a href="http://www.kclee.com/clemens/">
 *         Chr. Clemens Lee</a>
 *         &lt;<a href="mailto:clemens@kclee.com">
 *         clemens@kclee.com
 *         </a>>
 */
public class Init {
    private static final String S_PROJECTSUFFIX_KEY = "ProjectSuffix";
    private String S_INI_SUFFIX = ".ini";
    private String S_INIFILE_INIT          = "Init";
    private String S_INIFILE_INIT_FILE     = "File";
    private String S_INIFILE_INIT_OLDFILES = "OldFiles";
    private String S_INIFILE_INIT_DEBUG    = "Debug";
    private String S_INIFILE_INIT_AUTHOR   = "Author";
    private String S_INIFILE_HELP       = "Help";
    private String S_INIFILE_HELP_USAGE = "HelpUsage";
    private String S_INIFILE_COLORS          = "Colors";
    private String S_INIFILE_SYSTEM_COLORS   = "UseSystemColors";
    private String S_INIFILE_COLORS_BACK     = "Background";
    private String S_INIFILE_COLORS_LISTBACK = "ListBackground";
    private String S_INIFILE_COLORS_LISTFORE = "ListForeground";
    private String S_INIFILE_COLORS_LISTHIGHLIGHT = "ListHighlight";
    private String S_OPTION_STRING = " string";
   
    private int _max_projects = 6;
    private String _sInfoHeader;
   
    private Color _clrBackground = Color.gray;
    private Color _clrForeground = Color.black;
    private Color _clrListBackground = Color.gray;
    private Color _clrListForeground = Color.black;
    private Color _clrListHighlight = Color.black;
    private Color _clrTextBackground;
    private Color _clrTextForeground;
    private String _sAuthor = "<unknown author>";
    private String _sFileFullName = "";
    private String _sApplicationName;
    private String _sApplicationPath;
    private String _sFullIniFileName = null;
    public static final String S_FILE_SEPARATOR = System.getProperty("file.separator");
    private static final String S_PATH_SEPARATOR = System.getProperty("path.separator");
    
    private Hashtable _hshKeyValues = new Hashtable();
    
    private Hashtable _htOptions = new Hashtable();
    private Vector _vArguments = new Vector();
    
    /**
     * Reihenfolge der alten Files: ltestes File an Position 0
     */
    private Vector _vOldFiles = new Vector();

    /**
     * If no archive with the same name as the application
     * was found, "." is returned.
     */
    private String _getArchivePath(String sApplicationName_) {
        String sClasspath = System.getProperty("java.class.path");
        Vector vClasspath = Util.stringToLines
               (sClasspath, File.pathSeparatorChar);
        for(Enumeration eClasspath = vClasspath.elements();
            eClasspath.hasMoreElements(); )
        {
            String sNextPath = (String)eClasspath.nextElement();
            if (Util.endsWith(sNextPath, ".zip") ||
                Util.endsWith(sNextPath, ".jar"))
            {
                int lastSeparator = sNextPath.lastIndexOf
                       (File.separatorChar);
                String sArchiveName = sNextPath.substring
                       (lastSeparator + 1,
                        sNextPath.length() - 4);
                Util.debug("Init._getArchivePath(..).sArchiveName: " + sArchiveName);
                if (sArchiveName.startsWith(sApplicationName_)) {
                    String sApplicationPath = sNextPath.
                           substring(0, lastSeparator + 1);
                    if (Util.isEmpty(sApplicationPath)) {
                        sApplicationPath = ".";
                    }
                    return sApplicationPath;
                }
            }
        }
        
        return ".";
    }
    
    /**
     * in Classpath relative Pfade durch absolute ersetzen
     *
     * @deprecated   Use FileUtil.getAbsolutePathList(String) instead.
     */
    private void _setClassPath() {
        String sStartPath = System.getProperty("user.dir");
        Util.debug("_setClassPath: sStartPath: " + sStartPath);
        String sClassPath = System.getProperty("java.class.path");
        Util.debug("_setClassPath: sClassPath: " + sClassPath);
        Vector vClassPaths = Util.stringToLines(sClassPath,
                                                File.pathSeparatorChar);
        Util.debug("_setClassPath: vClassPaths: " + vClassPaths);
        for(int classPath = 0; classPath < vClassPaths.size();
            classPath++)
        {
            String sSingleClassPath = (String)vClassPaths.elementAt(classPath);
            Util.debug("_setClassPath: sSingleClassPath: " + sSingleClassPath);
            if (sSingleClassPath.startsWith(".")) {
                sSingleClassPath = FileUtil.concatPath(sStartPath,
                                                       sSingleClassPath);
                try {
                    File flSingleClassPath = new File(sSingleClassPath);
                    sSingleClassPath = flSingleClassPath.getCanonicalPath();
                } catch(Exception e) {
                }
                vClassPaths.setElementAt(sSingleClassPath, classPath);
            }
        }
        sClassPath = Util.concat(vClassPaths,
                                 File.pathSeparatorChar);
        if ((sClassPath.length() > 0) &&
            !Util.endsWith(sClassPath, File.pathSeparatorChar))
        {
            sClassPath += Util.cToS(File.pathSeparatorChar);
        }
        Util.debug("_setClassPath: sClassPath: " + sClassPath);
        Properties pProperties = System.getProperties();
        pProperties.put("java.class.path", sClassPath);
    }
   
    /**
     * This was the former Constructor.
     * Init colors and stuff.
     *
     * @see #<init>(Exitable, String[], String)
     */
    private void _init() {
        boolean bDebug = Util.atob(getKeyValue
                                   (S_INIFILE_INIT_DEBUG));
        if (bDebug) {
            Util.setDebug(bDebug);
        }
        
        // Hintergrund Farbe
        Vector vRGBString = null;
        {
            // user system colors?
            boolean bSystemColors = Util.atob
                   ( getKeyValue( S_INIFILE_COLORS,
                                  S_INIFILE_SYSTEM_COLORS ) );
            Util.debug( this, "_init().bSystemColors: " + bSystemColors );
            
            if ( bSystemColors ) {
                // is system unix and uses kde?
                if ( Util.isOSUnix() ) {
                    String sWindowManager = null;
                    try {
                        Process pProcess = 
                               Util.system( "set" );
                        BufferedReader pReader = new BufferedReader
                               ( new InputStreamReader( pProcess.getInputStream() ) );
                        sWindowManager  = pReader.readLine();
                    } catch( Exception pException ) {
                        sWindowManager = null;
                    }
                    Util.debug( this, "_init().sWindowManager: " +
                                sWindowManager );
                    //String sHomeDir = System.getProperty( "user.home" );
                    //String sKDEUserDir = FileUtil.concatPath( sHomeDir, ".kde"
                }
                _clrBackground = new Color( SystemColor.window.getRGB() );
                Util.debug( this, "_init().SystemColor: " + _clrBackground );
            } else {
                String sRGBBackground = getKeyValue(S_INIFILE_COLORS, S_INIFILE_COLORS_BACK);
                Util.debug("Init: sRGBBackground: " + sRGBBackground);
                vRGBString = Util.stringToLines(sRGBBackground, ' ');
                Util.debug("Init: vRGBString: " + vRGBString);
                if (vRGBString.size() == 3) {
                    _clrBackground = new Color(Util.atoi((String)vRGBString.elementAt(0)),
                                               Util.atoi((String)vRGBString.elementAt(1)),
                                               Util.atoi((String)vRGBString.elementAt(2)));
                } else {
                    /*_clrBackground = SystemColor.window;*/
                    _clrBackground = new Color(192, 192, 192);
                    Util.debug("Init: 0_clrBackground: " + _clrBackground);
                }
                Util.debug("Init: _clrBackground: " + _clrBackground);
            }
        }
        
        // ListHintergrund Farbe
        String sRGBColor = getKeyValue(S_INIFILE_COLORS,
                                       S_INIFILE_COLORS_LISTBACK);
        vRGBString = Util.stringToLines(sRGBColor, ' ');
        if (vRGBString.size() == 3) {
            _clrListBackground = new Color(Util.atoi((String)vRGBString.elementAt(0)),
                                           Util.atoi((String)vRGBString.elementAt(1)),
                                           Util.atoi((String)vRGBString.elementAt(2)));
        } else {
            _clrListBackground = Color.black;
        }
        // ListFordergrund Farbe
        sRGBColor = getKeyValue(S_INIFILE_COLORS, S_INIFILE_COLORS_LISTFORE);
        vRGBString = Util.stringToLines(sRGBColor, ' ');
        if (vRGBString.size() == 3) {
            _clrListForeground = new Color(Util.atoi((String)vRGBString.elementAt(0)),
                                           Util.atoi((String)vRGBString.elementAt(1)),
                                           Util.atoi((String)vRGBString.elementAt(2)));
        } else {
            _clrListForeground = new Color(192, 192, 192);
        }
        // List Highlight Color
        sRGBColor = getKeyValue(S_INIFILE_COLORS, S_INIFILE_COLORS_LISTHIGHLIGHT);
        vRGBString = Util.stringToLines(sRGBColor, ' ');
        if (vRGBString.size() == 3) {
            _clrListHighlight = new Color( Util.atoi( (String)vRGBString.elementAt( 0 ) ),
                                           Util.atoi( (String)vRGBString.elementAt( 1 ) ),
                                           Util.atoi( (String)vRGBString.elementAt( 2 ) ) );
        } else {
            _clrListHighlight = new Color(255, 255, 255);
        }
        _clrForeground = Color.black;
        _clrTextBackground = Color.white;
        _clrTextForeground = Color.black;
        
        /*// Language
          String sLanguageName = getKeyValue("Language");
          if (sLanguageName != null) {
          Language lngTemp = (Language)Language.get(sLanguageName);
          if (lngTemp != null) {
          _pLanguage = lngTemp;
          }
          }*/
        
        // FileName
        _sFileFullName = getKeyValue("File");
        Util.debug("Init: <init>: _sFileFullName: " + _sFileFullName);
        if (_sFileFullName == null) {
           _sFileFullName = "";
        }
        
        String sOldFiles = getKeyValue(S_INIFILE_INIT_OLDFILES);
        int i = 0;
        for (int j = 0; j < _max_projects; j++) {
            int k = sOldFiles.indexOf(59, i);
            if (k == -1)
                break;
            String string3 = sOldFiles.substring(i, k);
            _vOldFiles.addElement(new FileObject(string3));
            i = k + 1;
        }

        // author
        String tempAuthor = getKeyValue( S_INIFILE_INIT_AUTHOR );
        Util.debug( "_init().tempAuthor: " + tempAuthor );
        if ( !Util.isEmpty( tempAuthor ) ) {
            _sAuthor = tempAuthor;
        }
    }
    
    private boolean _bSurpressInifileOption = false;
    
    private String _sIniFileContent = null;
    
    /**
     * If sIniFileContent_ is null for the ini file will be first looked for in 
     * package/exitable.ini and if not found in package/package.ini.
     *
     * Following some implementation details:
     *
     * If the sIniFileContent_ parameter is set, this will
     * be used when reading ini content.
     *
     * All relative elements in the classpath will be exchanged
     * with their absolute counter parts.
     *
     * Ini file location:
     * 1. Set by user as option.
     */
    public Init(Exitable pExitable_,
                String[] asArg_, String sInfoHeader_,
                String sIniFileContent_)
    {
        super();

        _sIniFileContent = sIniFileContent_;
	Util.debug( this, "<init>(..)._sIniFileContent: " + _sIniFileContent );

        if (asArg_ == null) {
            asArg_ = new String[0];
        }

        _sInfoHeader = sInfoHeader_;
        
        String sClassName = pExitable_.getClass().getName();
        Util.debug("Init.<init>(..).sClassName: " + sClassName);
	 
        _setClassPath();
	 
        // Get the Package Name
        String sPackageName = ".";
        if(sClassName.indexOf('.') == -1) {
            _sApplicationName = new String(sClassName);
        } else {
            String sTempClassName = sClassName.substring
                    ( sClassName.lastIndexOf( '.' ) + 1 );
            if ( !sTempClassName.equals( "Main" ) &&
                 !sTempClassName.equals( "Controller" ) &&
                 !sTempClassName.equals( "ViewController" ))
            {
                sPackageName = sClassName.substring( 0, sClassName.lastIndexOf( '.' ) );
                _sApplicationName = sClassName.substring
                       ( sClassName.lastIndexOf( '.' ) + 1 );
            } else {
                sPackageName = sClassName.
                       substring(0, sClassName.lastIndexOf('.'));
                _sApplicationName = sPackageName.substring(sPackageName.lastIndexOf('.') + 1,
                                                           sPackageName.length());
                _sApplicationName = Util.firstCharToUpperCase(_sApplicationName);
            }
        }
        Util.debug( this, "<init>(..)._sApplicationName: " + _sApplicationName);
        
        // Get Application Path
        _sApplicationPath = _getArchivePath(_sApplicationName.toLowerCase());
        Util.debug( this, "<init>(..)._sApplicationPath.1: " +
                    _sApplicationPath );
        if (_sApplicationPath.equals(".")) {
            // no archive with same name found
            Util.debug( this, "<init>(..).sPackageName: " + sPackageName );
            _sApplicationPath = FileUtil.getPackagePath( sPackageName );
            Util.debug( this, "<init>(..)._sApplicationPath.2: " +
                        _sApplicationPath );
            if ( Util.isEmpty( _sApplicationPath ) ) {
                // another search for applicationname.jar
                _sApplicationPath = _getArchivePath( sPackageName );
                Util.debug( this, "<init>(..)._sApplicationPath.3: " +
                            _sApplicationPath );
            }
        }
        
        // First get options names and arguments
        Util.panicIf(asArg_ == null);
        for(int nr = 0; nr < asArg_.length; nr++) {
            Util.debug("Init.<init>(..).asArg_[" + nr + "]: " + asArg_[nr]);
            if (asArg_[nr].charAt(0) == '-') {
				// is it inifile option?
                if (asArg_[nr].equals("-inifile")) {
                    if (nr + 1 >= asArg_.length) {
                        Util.println("Error: No ini file name specified.");
                        pExitable_.setExit();
                        
                        return;
                    }
                    _sFullIniFileName = new String(asArg_[nr + 1]);
                    break;
                }
            }
        }
        
        // set full Ini file name
        if (Util.isEmpty(_sFullIniFileName)) {
            String sTempClassName = pExitable_.getClass().getName().toLowerCase();
            sTempClassName = sTempClassName.substring(sTempClassName.lastIndexOf('.') + 1);
            _sFullIniFileName = _sApplicationPath + File.separator +
                   sTempClassName + S_INI_SUFFIX;
            Util.debug("Init.<init>(..)._sFullIniFileName: " + _sFullIniFileName);
            if (FileUtil.existsFile(_sFullIniFileName)) {
                _sApplicationName = Util.firstCharToUpperCase(sTempClassName);
                _bSurpressInifileOption = true;
            } else {
                _sFullIniFileName = _sApplicationPath + File.separator +
                       _sApplicationName.toLowerCase() +
                       S_INI_SUFFIX;
                // hack ahead!!!
                // find /project/src/my/package/app.ini
                if ( !FileUtil.existsFile( _sFullIniFileName ) ) {
                    int classesIndex = _sFullIniFileName.lastIndexOf
                           ( File.separator + "classes" +
                             File.separator );
                    String sNewFullIniFileName = _sFullIniFileName.
                           substring( 0, classesIndex + 1 ) +
                           "src" + File.separator +
                           sPackageName.replace( '.', File.separatorChar ) +
                           File.separator + _sApplicationName.toLowerCase() +
                           S_INI_SUFFIX;
                    Util.debug( this, "<init>(..).sNewFullIniFileName: " +
                                sNewFullIniFileName );
                    if ( FileUtil.existsFile( sNewFullIniFileName ) ) {
                        _sFullIniFileName = sNewFullIniFileName;
                    }
                }
            }
        } else {
            _sFullIniFileName = FileUtil.getAbsoluteFileName(_sFullIniFileName);
            if (Util.isEmpty(_sFullIniFileName)) {
                Util.println("Error: No ini file name specified.");
                pExitable_.setExit();
                return;
            }
        }
        Util.debug("Init.<init>(..)._sFullIniFileName: " + _sFullIniFileName);
        
        // init colors and stuff
        _init();
        
        for(int nr = 0; nr < asArg_.length; nr++) {
            if (asArg_[nr].charAt(0) != '-') {
                _vArguments.addElement(new String(asArg_[nr]));
            } else {
                String sOption = asArg_[nr].substring(1, asArg_[nr].length());
                Util.debug("Init.<init>(..).sOption: " + sOption);
                if (sOption.equals("version")) {
                    Util.println( "" + _sApplicationName.toUpperCase() +
                                  " version " + getVersionString() +
                                  " " + getVersionDate() +
                                  " by " + getAuthor() );
                    pExitable_.setExit();
                } else if (sOption.equals("help")) {
                    printHelpMessage();
                    pExitable_.setExit();
                    return;
                } else if (sOption.equals("inifile")) {
                    nr++;
                    if (_bSurpressInifileOption) {
                        Util.println("Error: No custom ini file supported!");
                        pExitable_.setExit();
                        
                        return;
                    }
                } else if (sOption.equals("debug")) {
                    Util.setDebug(true);
                } else {
                    // user option
                    String sOptionDocu = getKeyValue(S_INIFILE_HELP,
                                                     sOption);
                    // Does the option really exists?
                    if (!Util.isEmpty(sOptionDocu)) {
                        Option pOption = new Option(sOption, sOptionDocu);
                        Object oOptionValue = new Boolean(true);
                        if (pOption.getType() == Option.STRING) {
                            if (nr + 1 >= asArg_.length) {
                                Util.println("Error: Option " + sOption + " needs a string value.");
                                printHelpMessage();
                                pExitable_.setExit();
                                return;
                            }
                            oOptionValue = new String(asArg_[nr + 1]);
                            nr++;
                        }
                        _htOptions.put(sOption, oOptionValue);
                    } else {
                        Util.println("Error: Option " + sOption + " ist not acceptable.");
                        printHelpMessage();
                        pExitable_.setExit();
                        return;
                    }
                }
            }
        }
    }
    
    /**
     * The ini file will be first looked for in package/exitable.ini
     * and if not found in package/package.ini.
     *
     */
    public Init(Exitable pExitable_,
                String[] asArg_, String sInfoHeader_)
    {
        this(pExitable_, asArg_, sInfoHeader_, null);
    }
    
   public void printHelpMessage() {
      String sOptions = getKeyValue(S_INIFILE_HELP, "Options");
      Vector vOptions = Util.stringToLines(sOptions, ',');
      // htOptions enthlt Option Objekte
      Hashtable htOptions = new Hashtable();
      // zum formatieren der help messages
      int maxOptionLength = 7;
      for(Enumeration eOptions = vOptions.elements();
          eOptions.hasMoreElements(); )
      {
          String sNextOption = (String)eOptions.nextElement();
	  Util.debug( this, "printHelpMessage().sNextOption: " + sNextOption );
          Option pOption = new Option(sNextOption,
                                      getKeyValue(S_INIFILE_HELP,
                                                  sNextOption));
          
          htOptions.put(sNextOption, pOption);
          
          if (sNextOption.length() > maxOptionLength) {
              maxOptionLength = sNextOption.length();
          }
      }
      
      String sUsage = getKeyValue(S_INIFILE_HELP,
											 S_INIFILE_HELP_USAGE);
      String sSynopsis = "Usage: " + _sApplicationName.toLowerCase() + " ";
      String sMandatory = "";
      String sOptional = "";
      String sRemarks ="";
      for(Enumeration eOptions = vOptions.elements();
			 eOptions.hasMoreElements(); )
      {
			String sOption = (String)eOptions.nextElement();
			Option optNext = (Option)htOptions.get(sOption);
			sRemarks +=  "       " + sOption + ":   " +
					 Util.getSpaces(maxOptionLength -
										 sOption.length()) +
					 optNext.getDescription() + "\n";
			if (optNext.getType() == Option.STRING) {
				sOption += S_OPTION_STRING;
			}
			if (optNext.isMandatory()) {
				sMandatory += "-" + sOption + " ";
			} else {
				sOptional += "[-" + sOption + "] ";
			}
      }
      sRemarks += "       version:   " +
				 Util.getSpaces(maxOptionLength - 7) +
				 "Prints the version number of the program.\n" +
				 "       help:   " +
				 Util.getSpaces(maxOptionLength - 4) +
				 "Prints this help message.\n" +
				 "       debug:   " +
				 Util.getSpaces(maxOptionLength - 5) +
				 "Prints debugging information while running.";
		if (!_bSurpressInifileOption) {
			sRemarks += "\n       inifile:   " +
					 Util.getSpaces(maxOptionLength - 7) +
					 "Starts this application with another ini file than\n" +
					 "           " +
					 Util.getSpaces(maxOptionLength) +
					 "the default one.";
		}
      sOptional += "[-version] [-help] [-debug] [-inifile foo.ini] ";
      sSynopsis += sMandatory + sOptional + sUsage;
      Util.println(sSynopsis);
      Util.println(sRemarks);
   }
   
   /**
     * @return true on error.
     */
   public synchronized boolean setKeyValue(String sKey_, int value_) {
      return setKeyValue(sKey_, Util.itoa(value_));
   }
   
   /**
     * @return true on error.
     */
   public synchronized boolean setKeyValue(String sKey_,
					   String sValue_)
      {
	 return setKeyValue(S_INIFILE_INIT, sKey_, sValue_);
      }
   
    /**
     * @return true on error.
     */
    public synchronized boolean setKeyValue(String sSection_,
                                            String sKey_,
                                            String sValue_)
    {
        _hshKeyValues.put(sSection_ + sKey_, sValue_);
        IniFile.setKeyValue(_sFullIniFileName, sSection_,
                            sKey_, sValue_);
        return(IniFile.getStatus() != IniFile.OK);
    }
   
   public Vector getArguments() {
      return (Vector)_vArguments.clone();
   }
   
   public Hashtable getOptions() {
      return (Hashtable)_htOptions.clone();
   }
   
   public Enumeration getArgumentsElements() {
      return(_vArguments.elements());
   }
   
    public synchronized String getKeyValue(String sSection_,
                                           String sKey_)
    {
        if (_hshKeyValues.containsKey(sSection_ + sKey_)) {
            return (String)_hshKeyValues.get(sSection_ + sKey_);
        }
        String sValue = IniFile.getKeyValue(_sFullIniFileName,
                                            sSection_, sKey_,
                                            _sIniFileContent);
        Util.debug("Init.getKeyValue(..).IniFile.status(): " +
                   IniFile.getStatus());
        _hshKeyValues.put(sSection_ + sKey_, sValue);
        
        return sValue;
    }
   
   public synchronized String getKeyValue(String sKey_) {
      return getKeyValue(S_INIFILE_INIT, sKey_);
   }
   
   /**
     * @see #getVersionString
     */
   public String getInfoHeader() {
      return(_sInfoHeader);
   }
   
   public Color getBackground() {
      return _clrBackground;
   }
   
   public Color getForeground() {
      return _clrForeground;
   }
   
   public Color getListBackground() {
      return _clrListBackground;
   }
   
   public Color getListForeground() {
      return _clrListForeground;
   }
   
   public Color getListHighlight() {
      return new Color( _clrListHighlight.getRGB() );
   }
   
   public String colorToString(Color pColor_) {
      Util.panicIf(pColor_ == null);
      
      return("" + pColor_.getRed() + " " +
	     pColor_.getGreen() + " " + pColor_.getBlue());
   }
   
   public void setBackground(Color pColor_) {
      Util.panicIf(pColor_ == null);
      _clrBackground = new Color(pColor_.getRGB());
      setKeyValue(S_INIFILE_COLORS, S_INIFILE_COLORS_BACK,
		  colorToString(_clrBackground));
   }
   
   public void setListBackground(Color pColor_) {
      Util.panicIf(pColor_ == null);
      _clrListBackground = new Color(pColor_.getRGB());
      setKeyValue(S_INIFILE_COLORS, S_INIFILE_COLORS_LISTBACK,
		  colorToString(_clrListBackground));
   }
   
   public void setListForeground(Color pColor_) {
      Util.panicIf(pColor_ == null);
      _clrListForeground = new Color(pColor_.getRGB());
      setKeyValue(S_INIFILE_COLORS, S_INIFILE_COLORS_LISTFORE,
		  colorToString(_clrListForeground));
   }
   
   public void setListHighlight(Color pColor_) {
      Util.panicIf(pColor_ == null);
      _clrListHighlight = new Color(pColor_.getRGB());
      setKeyValue(S_INIFILE_COLORS, S_INIFILE_COLORS_LISTHIGHLIGHT,
		  colorToString(_clrListHighlight));
   }
   
    /**
     * The first char is upper case (like "Project").
     */
    public String getApplicationName() {
        return _sApplicationName;
    }

    /**
     * @return   for example: "/home/clemens/src/java/jacob"
     */
    public String getApplicationPath() {
        return _sApplicationPath;
    }

    private void _saveOldFiles() {
        // Alte Files speichern
        String sOldFiles = "";
        for (int i = 0; i < _vOldFiles.size(); i++) {
            FileObject pFileObject = (FileObject)_vOldFiles.elementAt(i);
            Util.debug( "Init._saveOldFiles()._vOldFiles: " +
                        _vOldFiles );
            if ( Util.equalsCaseless(pFileObject.getName(),
                                     getFileName()) )
            {
                _vOldFiles.removeElementAt(i);
                i--;
            } else {
                sOldFiles += pFileObject.getFileFullName() + ";";
            }
        }
        if (Util.getOccurances(sOldFiles, 59) > _max_projects) {
            sOldFiles = sOldFiles.substring(sOldFiles.indexOf(59) + 1, sOldFiles.length());
            _vOldFiles.removeElementAt(0);
        }
        Util.debug( "Init._saveOldFiles().sOldFiles: " + sOldFiles );
        setKeyValue(S_INIFILE_INIT_OLDFILES, sOldFiles);
    }

    /**
     * For File->New use "" or null for sFileFullName_.
     *
     * @return   IniFile status.
     */
   public boolean setFileFullName(String sFileFullName_) {
       if ( !Util.isEmpty( getFileFullName() ) ) {
           makeThisFileOld();
       }

       if ( Util.isEmpty( sFileFullName_ ) ) {
           _sFileFullName = "";
           setKeyValue( S_INIFILE_INIT, S_INIFILE_INIT_FILE,
                        "" );
           _saveOldFiles();

           int status = IniFile.getStatus();
           boolean bRetVal = (status != IniFile.OK);

           return bRetVal;
       }

      // falls relativer Pfad -> absolut
      _sFileFullName = FileUtil.getAbsoluteFileName(sFileFullName_);
      Util.debug("Init: setFileFullName: _sApplicationPath: " + _sApplicationPath);
      Util.debug("Init: setFileFullName: _sApplicationName: " + _sApplicationName);
      setKeyValue(S_INIFILE_INIT,
                  S_INIFILE_INIT_FILE, _sFileFullName);
      int status = IniFile.getStatus();
      boolean bRetVal = (status != IniFile.OK);
      /*AWTUtil.showMessage("Error: " + Util.firstCharToLowerCase(_sApplicationName) +
		  S_INI_SUFFIX + " File could not be written to disk!");*/
      
      // Alte Files speichern
      Util.debug("Init: setFileFullName: _vOldFiles: " +
                 _vOldFiles);
      String sOldFiles = "";
      for (int i = 0; i < _vOldFiles.size(); i++) {
          FileObject pFileObject = (FileObject)_vOldFiles.elementAt(i);
          if ( Util.equalsCaseless(pFileObject.getName(),
                                   getFileName()) )
          {
              Util.debug("Init: setFileFullName: pFileObject.getName(): " +
                         pFileObject.getName());
              Util.debug("Init.setFileFullName().getFileName(): " +
                         getFileName());
              _vOldFiles.removeElementAt(i);
              i--;
          } else {
              sOldFiles += pFileObject.getFileFullName() + ";";
          }
      }
      if (Util.getOccurances(sOldFiles, 59) > _max_projects) {
          sOldFiles = sOldFiles.substring(sOldFiles.indexOf(59) + 1, sOldFiles.length());
          _vOldFiles.removeElementAt(0);
      }
      Util.debug("Init: setFileFullName: sOldFiles: " + sOldFiles);
      setKeyValue(S_INIFILE_INIT_OLDFILES, sOldFiles);
      
      return bRetVal;
   }
	
   /**
    * First char is upper case, the rest lower case. Only the name
    * is returned, no path or file postfix information is given.
    *
    * @see #getFileFullName
    * @see #getFilePath
    */
   public String getFileName() {
      // \ zu / umtauschen
      Util.debug("Init: getFileName: _sFileFullName: " + _sFileFullName);
      String sRetVal = new String(_sFileFullName);
      sRetVal = sRetVal.replace('\\', '/');
      
      // String zwische letztem / und .
      int indexStart = sRetVal.lastIndexOf('/') + 1;
      int indexStop  = sRetVal.lastIndexOf('.');
      if (indexStart == -1 || indexStop == -1 || 
			 indexStart > indexStop) 
		{
			return "";
      }
      sRetVal = sRetVal.substring(indexStart, indexStop);
      sRetVal = Util.firstCharToUpperCase(sRetVal);
      
      return sRetVal;
   }
   
   public String getFileFullName() {
      return _sFileFullName;
   }
   
   /**
     * @return File separator is '/' and the path ends with a file
     *         separator char.
     */
   public String getFilePath() {
      // \ zu / umtauschen
      String sRetVal = new String(_sFileFullName);
      sRetVal = sRetVal.replace('\\', '/');
      
      // String zwische letztem / und .
      int indexStop = sRetVal.lastIndexOf('/') + 1;
      Util.panicIf(indexStop == -1);
      sRetVal = sRetVal.substring(0, indexStop);
      
      return sRetVal;
   }
   
   public Enumeration getOldFilesElements() {
      return _vOldFiles.elements();
   }
   
   public int getOldFilesSize() {
      return _vOldFiles.size();
   }
   
   public String getOldFileFullName(String string) {
      for (Enumeration enumeration = _vOldFiles.elements();
	   enumeration.hasMoreElements(); )
      {
	 FileObject pFileObject = (FileObject)enumeration.nextElement();
	 if (Util.equalsCaseless(pFileObject.getName(), string))
	    return pFileObject.getFileFullName();
      }
      Util.panicIf(true, "Init: getOldFileFullName: Muss vorher was gefunden haben.");
      return null;
   }
   
    /**
     * This method stores the actual file in the old file vector,
     * no saving or anything else like removing double fils.
     */
    public void makeThisFileOld() {
        if (_sFileFullName != null && (!_sFileFullName.equals(""))) {
            Util.debug( "Init.makeThisFileOld()._sFileFullName: " +
                        _sFileFullName );
            FileObject pFileObject = new FileObject( _sFileFullName );
            Util.debug( "Init.makeThisFileOld().pFileObject: " +
                        pFileObject );
            _vOldFiles.addElement( pFileObject );
            Util.debug("Init.makeThisFileOld()._vOldFiles: " +
                       _vOldFiles);
        }
    }
   
    /*public Language getLanguage() {
      return _pLanguage;
      }*/
   
   public String getVersionString() {
      Util.debug( this, "getVersionString()._sInfoHeader: " + _sInfoHeader );
      if (_sInfoHeader == null) {
			return "1.1";
      }
      int index = _sInfoHeader.indexOf(" ", 9) + 1;
      String sHead = _sInfoHeader.substring(index, _sInfoHeader.length());
      index = sHead.indexOf(' ');
      String sVersion = sHead.substring(0, index);
      Util.debug( this, "getVersionString().sVersion: " + sVersion );

      return sVersion;
   }
   
   public int getVersion() {
      String sVersion = getVersionString();
      sVersion = sVersion.substring(0, sVersion.indexOf('.'));
      
      return Util.atoi(sVersion);
   }
   
   public int getRelease() {
      String sVersion = getVersionString();
      String sRelease = sVersion.substring(sVersion.indexOf('.') + 1,
					   sVersion.length());
      
      return Util.atoi(sRelease);
   }
   
   /**
     * @return The right language is automaticly used.
     */
   public String getVersionDate() {
      String sVersion = getVersionString();
      int index = _sInfoHeader.indexOf(sVersion) + sVersion.length() + 1;
      String sHead = _sInfoHeader.substring(index, _sInfoHeader.length());
      index = sHead.indexOf(' ');
      String sDate = sHead.substring(0, index);
      
      // Format Date
      int year = Util.atoi(sDate.substring(0, 4));
      int month = Util.atoi(sDate.substring(5, 7)) - 1;
      int day = Util.atoi(sDate.substring(8, 10));
      /*MultiDate pMultiDate = new MultiDate(year, month, day,
        getLanguage());
        sDate = pMultiDate.toLocaleDateString();*/
      sDate = "" + year + "-" + Util.paddWithZero
             ( month + 1, 2 ) + "-" + Util.paddWithZero( day, 2 );
      
      return sDate;
   }
   
   public String getVersionTime() {
      String sVersion = getVersionString();
      int index = _sInfoHeader.indexOf(sVersion) + sVersion.length() + 12;
      String sHead = _sInfoHeader.substring(index, _sInfoHeader.length());
      index = sHead.indexOf(' ');
      String sTime = sHead.substring(0, index);
      
      return sTime;
   }

    public static String getProjectSuffixKey() {
        return S_PROJECTSUFFIX_KEY;
    }

    public String getAuthor() {
        return _sAuthor;
    }

    public void setAuthor( String sAuthor_ ) {
        _sAuthor = sAuthor_;
    }
}
